/* -*-c-*-
 *
 * Prime fields with efficient reduction for special-form primes
 *
 * (c) 2004 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/sub.h>

#include "field.h"
#include "field-guts.h"
#include "mprand.h"

/*----- Main code ---------------------------------------------------------*/

/* --- Field operations --- */

static void fdestroy(field *ff) {
  fctx_niceprime *f = (fctx_niceprime *)ff;
  mpreduce_destroy(&f->r);
  DESTROY(f);
}

static mp *frand(field *ff, mp *d, grand *r) {
  fctx_niceprime *f = (fctx_niceprime *)ff;
  return (mprand_range(d, f->r.p, r, 0));
}

static int fzerop(field *ff, mp *x) { return (MP_ZEROP(x)); }

static mp *fneg(field *ff, mp *d, mp *x) {
  fctx_niceprime *f = (fctx_niceprime *)ff;
  return (mp_sub(d, f->r.p, x));
}

static mp *fadd(field *ff, mp *d, mp *x, mp *y) {
  fctx_niceprime *f = (fctx_niceprime *)ff; d = mp_add(d, x, y);
  if (MP_NEGP(d)) d = mp_add(d, d, f->r.p);
  else if (MP_CMP(d, >, f->r.p)) d = mp_sub(d, d, f->r.p);
  return (d);
}

static mp *fsub(field *ff, mp *d, mp *x, mp *y) {
  fctx_niceprime *f = (fctx_niceprime *)ff; d = mp_sub(d, x, y);
  if (MP_NEGP(d)) d = mp_add(d, d, f->r.p);
  else if (MP_CMP(d, >, f->r.p)) d = mp_sub(d, d, f->r.p);
  return (d);
}

static mp *fmul(field *ff, mp *d, mp *x, mp *y) {
  fctx_niceprime *f = (fctx_niceprime *)ff; d = mp_mul(d, x, y);
  return (mpreduce_do(&f->r, d, d));
}

static mp *fsqr(field *ff, mp *d, mp *x) {
  fctx_niceprime *f = (fctx_niceprime *)ff; d = mp_sqr(d, x);
  return (mpreduce_do(&f->r, d, d));
}

static mp *finv(field *ff, mp *d, mp *x) {
  fctx_niceprime *f = (fctx_niceprime *)ff;
  d = mp_modinv(d, x, f->r.p);
  return (d);
}

static mp *freduce(field *ff, mp *d, mp *x) {
  fctx_niceprime *f = (fctx_niceprime *)ff;
  return (mpreduce_do(&f->r, d, x));
}

static mp *fsqrt(field *ff, mp *d, mp *x) {
  fctx_niceprime *f = (fctx_niceprime *)ff;
  return (mp_modsqrt(d, x, f->r.p));
}

static mp *fdbl(field *ff, mp *d, mp *x) {
  fctx_niceprime *f = (fctx_niceprime *)ff; d = mp_lsl(d, x, 1);
  if (MP_CMP(d, >=, f->r.p)) d = mp_sub(d, d, f->r.p);
  return (d);
}

static mp *ftpl(field *ff, mp *d, mp *x) {
  fctx_niceprime *f = (fctx_niceprime *)ff; MP_DEST(d, MP_LEN(x) + 1, x->f);
  MPX_UMULN(d->v, d->vl, x->v, x->vl, 3); d->f &= ~MP_UNDEF;
  while (MP_CMP(d, >=, f->r.p)) d = mp_sub(d, d, f->r.p);
  return (d);
}

static mp *fqdl(field *ff, mp *d, mp *x) {
  fctx_niceprime *f = (fctx_niceprime *)ff; d = mp_lsl(d, x, 2);
  while (MP_CMP(d, >=, f->r.p)) d = mp_sub(d, d, f->r.p);
  return (d);
}

static mp *fhlv(field *ff, mp *d, mp *x) {
  fctx_niceprime *f = (fctx_niceprime *)ff;
  if (MP_ZEROP(x)) { MP_COPY(x); MP_DROP(d); return (x); }
  if (x->v[0] & 1) { d = mp_add(d, x, f->r.p); x = d; }
  return (mp_lsr(d, x, 1));
}

/* --- Field operations table --- */

static const field_ops fops = {
  FTY_PRIME, "niceprime",
  fdestroy, frand, field_stdsamep,
  freduce, field_id,
  fzerop, fneg, fadd, fsub, fmul, fsqr, finv, freduce, fsqrt,
  0,
  fdbl, ftpl, fqdl, fhlv
};

/* --- @field_niceprime@ --- *
 *
 * Arguments:	@mp *p@ = the characteristic of the field
 *
 * Returns:	A pointer to the field, or null.
 *
 * Use:		Creates a field structure for a prime field of size %$p$%,
 *		using efficient reduction for nice primes.
 */

field *field_niceprime(mp *p)
{
  fctx_niceprime *f = CREATE(fctx_niceprime);
  f->f.ops = &fops;
  f->f.zero = MP_ZERO;
  f->f.one = MP_ONE;
  f->f.nbits = mp_bits(p);
  f->f.noctets = (f->f.nbits + 7) >> 3;
  if (mpreduce_create(&f->r, p)) {
    DESTROY(f);
    return (0);
  }
  f->f.m = f->r.p;
  f->f.q = f->r.p;
  return (&f->f);
}

/*----- That's all, folks -------------------------------------------------*/
